/***********************************************************************/
/*                                                                     */
/*   svm_learn.c                                                       */
/*                                                                     */
/*   Learning module of Support Vector Machine.                        */
/*                                                                     */
/*   Author: Thorsten Joachims                                         */
/*   Date: 06.07.98                                                    */
/*                                                                     */
/*   Copyright (c) 1998  Universitaet Dortmund - All rights reserved   */
/*                                                                     */
/*   This software is available for non-commercial use only. It must   */
/*   not be modified and distributed without prior permission of the   */
/*   author. The author is not responsible for implications from the   */
/*   use of this software.                                             */
/*                                                                     */
/***********************************************************************/

# include "svm_common.h"
 
char docfile[200];           /* file with training examples */
char modelfile[200];         /* file for resulting classifier */

double epsilon_crit;         /* tolerable error for distances used */
                             /* in stopping criterion */
double epsilon_shrink;       /* how much a multiplier should be above */
                             /* zero for shrinking */
double epsilon_a;            /* tolerable error on alphas */
double epsilon_const;        /* tolerable error on eq-constraint */

long   *kernel_cache_index;  /* cache some kernel evalutations */
CFLOAT *kernel_cache_buffer; /* to improve speed */
long   *kernel_cache_invindex;
long   *kernel_cache_active2totdoc;
long   *kernel_cache_totdoc2active;
long   *kernel_cache_lru;
long   *kernel_cache_occu;
long   kernel_cache_elems=0;
long   max_kernel_cache_elems=0;
long   kernel_cache_time=0;
long   kernel_cache_malloc();
long   kernel_cache_activenum;
long   kernel_cache_buffsize;

double *optimize_qp();       /* interface to QP-solver */
double *opt_ce,*opt_ce0,*opt_g,*opt_g0,*opt_xinit; 
long   opt_n;
double opt_precision=1e-14;  /* precision of solver, set to 1e-21 */
                             /* if you get convergence problems */
double svm_c;
int nx;


void   svm_learn(DOC *, long *, long, long, long, long, long, double, 
		 KERNEL_PARM *, MODEL *);
double compute_objective_function(double *, double *, long *, long *);
long   compute_index(long *,long, long *);
void   optimize_svm(DOC *, long *, long *, long *, MODEL *, long, long *, long,
		    double *, double *, double, CFLOAT *, KERNEL_PARM *);
void   compute_matrices_for_optimization(DOC *, long *, long *, long *, 
					 long *, MODEL *, double *, double *,
					 long, long, double, CFLOAT *,
					 KERNEL_PARM *);
long   calculate_svm_model(DOC *, long *, double *, double *, double, long *,
			   long, long *, MODEL *, KERNEL_PARM *);
long   check_optimality(MODEL *, long *, double *, double *,long, double,
			double *, long *, long *, long *, long *, long);
long   identify_inconsistent(double *, long, double, long *, long *);
void   update_linear_component(DOC *, long *, long *, double *, double *, 
			       long *, long, KERNEL_PARM *, double *, double *,
			       CFLOAT *);
long   select_next_qp_subproblem_grad(long *, double *, double *, long, long,
				      double, long *, long *, long, long *,
				      double *, long *, long *, long *);
long   select_next_qp_subproblem_grad_cache(long *, double *, double *, long, 
					    long, double, long *, long *, 
					    long, long *, double *, long *,
					    long *, long *);
void   select_top_n(double *, long, long *, long);
long   shrink_problem(long, long *, long *,long *, long *, long *, long,
		      long, long, double *, double **, long *);
void   reactivate_inactive_examples(long *, double *, double **, double *, 
				    double *, long, long, double, long *, 
				    long *, long *, long, DOC *, KERNEL_PARM *,
				    MODEL *, CFLOAT *);
void   get_kernel_row(DOC *, long, long, long *, CFLOAT *, KERNEL_PARM *);
void   cache_kernel_evals(DOC *, long *, long, KERNEL_PARM *);
void   kernel_cache_shrink(long, long, long *);
void   kernel_cache_init(long, long);
void   kernel_cache_cleanup();
long   kernel_cache_malloc();
void   kernel_cache_free(long);
long   kernel_cache_free_lru();
CFLOAT *kernel_cache_clean_and_malloc(long);
long   kernel_cache_touch(long);
long   kernel_cache_check(long);
double estimate_margin_vcdim(MODEL *, double, KERNEL_PARM *);
double estimate_sphere(MODEL *, KERNEL_PARM *);
void   write_model(char *, MODEL *, KERNEL_PARM *);
void   read_input_parameters(int, char **, char *, char *,long *, long *, 
			     long *, long *, long *, long *, double *, 
			     double *, KERNEL_PARM *);
void   print_help();

/*---------------------------------------------------------------------------*/

void main (argc, argv)
int argc;
char *argv[];
{  
  DOC *docs;  /* training examples */
  long *label,max_docs,max_words_doc,svm_maxqpsize,svm_iter_to_shrink;
  long totwords,totdoc,ll,remove_inconsistent,skip_final_opt_check;
  long kernel_cache_size;
  KERNEL_PARM kernel_parm;
  MODEL model;

  read_input_parameters(argc,argv,docfile,modelfile,&verbosity,
			&remove_inconsistent,&skip_final_opt_check,
			&svm_maxqpsize,&svm_iter_to_shrink,&kernel_cache_size,
			&svm_c,&epsilon_crit,&kernel_parm);

  epsilon_a = opt_precision*10000; /* tolerable error on alpha=0 or alpha=C */

  nol_ll(docfile,&max_docs,&max_words_doc,&ll); /* scan size of input file */
  max_words_doc+=2;
  ll+=2;
  max_docs+=2;

  docs = (DOC *)my_malloc(sizeof(DOC)*max_docs);
  label = (long *)my_malloc(sizeof(long)*max_docs);

  read_documents(docfile,docs,label,max_words_doc,ll,&totwords,&totdoc);

  kernel_cache_init(totdoc,kernel_cache_size);
  svm_learn(docs,label,totdoc,remove_inconsistent,skip_final_opt_check,
	    svm_maxqpsize,svm_iter_to_shrink,svm_c,&kernel_parm,&model);
  kernel_cache_cleanup();

  write_model(modelfile,&model,&kernel_parm);

  free(model.supvec);
  free(model.alpha);
  free(docs);
  free(label);
}

/*---------------------------------------------------------------------------*/

void svm_learn(docs,label,totdoc,remove_inconsistent,skip_final_opt_check,
	       svm_maxqpsize,svm_iter_to_shrink,svm_c,kernel_parm,model)
DOC *docs;
long *label,totdoc,remove_inconsistent,skip_final_opt_check,svm_maxqpsize;
long svm_iter_to_shrink;
double svm_c;
KERNEL_PARM *kernel_parm;
MODEL *model;
{
  long *inconsistent,*chosen,*key,i,jj,*last_suboptimal_at;
  long inconsistentnum,choosenum,retrain,iteration,inactivenum;
  long misclassified,supvecnum=0,upsupvecnum,qp_size,*active2dnum;
  long at_upper_bound,*active,*working2dnum,varnum,*selexam;
  long activenum,*inactive_since,deactnum;
  double model_length,criterion,eq;
  double maxdiff,*a_old,*lin,*a,*ub_lin;
  double **a_history;
  long t0=0,t1=0,t2=0,t3=0,t4=0,t5=0,t6=0; /* timing */
  long time_kernel,time_opti,time_shrink,time_update;
  long runtime_start,runtime_end;
  long time_model,time_check,time_select;
  double *selcrit;  /* buffer for sorting */        
  CFLOAT *aicache;  /* buffer to keep one row of hessian */

  runtime_start=get_runtime();
  time_kernel=0;
  time_opti=0;
  time_shrink=0;
  time_update=0;
  time_model=0;
  time_check=0;
  time_select=0;
  kernel_cache_statistic=0;

  epsilon_shrink=2;
  maxdiff=1;

  inconsistent = (long *)my_malloc(sizeof(long)*totdoc);
  chosen = (long *)my_malloc(sizeof(long)*totdoc);
  active = (long *)my_malloc(sizeof(long)*totdoc);
  inactive_since = (long *)my_malloc(sizeof(long)*totdoc);
  last_suboptimal_at = (long *)my_malloc(sizeof(long)*totdoc);
  key = (long *)my_malloc(sizeof(long)*totdoc);
  selcrit = (double *)my_malloc(sizeof(double)*totdoc);
  selexam = (long *)my_malloc(sizeof(long)*totdoc);
  a = (double *)my_malloc(sizeof(double)*totdoc);
  a_old = (double *)my_malloc(sizeof(double)*totdoc);
  a_history = (double **)my_malloc(sizeof(double *)*1000);
  lin = (double *)my_malloc(sizeof(double)*totdoc);
  ub_lin = (double *)my_malloc(sizeof(double)*totdoc);
  aicache = (CFLOAT *)my_malloc(sizeof(CFLOAT)*totdoc);
  working2dnum = (long *)my_malloc(sizeof(long)*(totdoc+1));
  active2dnum = (long *)my_malloc(sizeof(long)*(totdoc+1));
  model->supvec = (DOC **)my_malloc(sizeof(DOC *)*(totdoc+2));
  model->alpha = (double *)my_malloc(sizeof(double)*(totdoc+2));
  opt_ce = (double *)my_malloc(sizeof(double)*svm_maxqpsize);
  opt_ce0 = (double *)my_malloc(sizeof(double));
  opt_g = (double *)my_malloc(sizeof(double)*svm_maxqpsize*svm_maxqpsize);
  opt_g0 = (double *)my_malloc(sizeof(double)*svm_maxqpsize);
  opt_xinit = (double *)my_malloc(sizeof(double)*svm_maxqpsize);

  qp_size=svm_maxqpsize;
  model->sv_num=0;
  activenum=totdoc;
  choosenum=0;
  inactivenum=0;
  deactnum=0;
  inconsistentnum=0;

  for(i=0;i<totdoc;i++) {    /* various inits */
    inconsistent[i]=0;
    chosen[i]=0;
    active[i]=1;
    a[i]=0;
    a_old[i]=0;
    lin[i]=0;
    ub_lin[i]=0;
    last_suboptimal_at[i]=1;
  }
  retrain=1;
  activenum=compute_index(active,totdoc,active2dnum);
  varnum=compute_index(chosen,totdoc,working2dnum);

  if(verbosity==1) {
    printf("Optimizing"); fflush(stdout);
  }

                            /* repeat this loop until we have convergence */
  for(iteration=1;retrain;iteration++) {

    kernel_cache_time=iteration;  /* for lru cache */
    if(verbosity>=2) {
      printf(
	"Iteration %ld: ",iteration); fflush(stdout);
    }
    else if(verbosity==1) {
      printf("."); fflush(stdout);
    }

    if(verbosity>=2) t0=get_runtime();
    if(verbosity>=3) {
      printf("\nSelecting working set... "); fflush(stdout); 
    }
    choosenum=0;
    for(jj=0;working2dnum[jj]>=0;jj++) {
      i=working2dnum[jj];
      chosen[i]=0;  
    }
    if(retrain == 2) {
      for(i=0;i<totdoc;i++) { /* set inconsistent examples to zero (-i 1) */
	if(inconsistent[i] && (a[i] != 0.0)) {
	  chosen[i]=1;
	  choosenum++;
	  a[i]=0;
	}
      }
      eq=0;
      for(i=0;i<totdoc;i++) { /* make sure we fulfill equality constraint */
	eq+=a[i]*label[i];
      }
      for(i=0;(i<totdoc) && (eq != 0);i++) {
	if(eq*label[i] > 0) {
	  chosen[i]=1;
	  choosenum++;
	  if((eq*label[i]) > a[i]) {
	    eq-=(a[i]*label[i]);
	    a[i]=0;
	  }
	  else {
	    a[i]-=(eq*label[i]);
	    eq=0;
	  }
	}
      }
    }
    else if(0) {  /* select working set according to steepest gradient */ 
      choosenum=select_next_qp_subproblem_grad(label,a,lin,totdoc,
					       qp_size,svm_c,
					       inconsistent,active2dnum,varnum,
					       working2dnum,selcrit,selexam,key,
					       chosen);
    }
    else if(1) {  /* select half of the working set from cache */
      if(qp_size>=4) {
	choosenum=select_next_qp_subproblem_grad_cache(label,a,lin,totdoc,
					       (long)(qp_size/2),svm_c,
					       inconsistent,active2dnum,varnum,
					       working2dnum,selcrit,selexam,
					       key,chosen);
      }
      else {
	choosenum=0;
      }
      choosenum+=select_next_qp_subproblem_grad(label,a,lin,totdoc,
                                             (long)(qp_size-choosenum),svm_c,
					     inconsistent,active2dnum,varnum,
					     working2dnum,selcrit,selexam,key,
					     chosen);
    }

    if(verbosity>=2) {
      printf(" %ld vectors chosen\n",choosenum); fflush(stdout); 
    }

    varnum=compute_index(chosen,totdoc,working2dnum);

    if(verbosity>=2) t1=get_runtime();
    cache_kernel_evals(docs,working2dnum,varnum,kernel_parm);

    if(verbosity>=2) t2=get_runtime();
    if(retrain != 2) {
      optimize_svm(docs,label,chosen,active2dnum,model,totdoc,working2dnum,
		   varnum,a,ub_lin,svm_c,aicache,kernel_parm);
      epsilon_a = opt_precision * 10000; 
    }

    if(verbosity>=2) t3=get_runtime();
    update_linear_component(docs,label,active2dnum,a,a_old,working2dnum,totdoc,
			    kernel_parm,lin,ub_lin,aicache);

    if(verbosity>=2) t4=get_runtime();
    supvecnum=calculate_svm_model(docs,label,a,lin,svm_c,chosen,totdoc,
				  &at_upper_bound,model,kernel_parm);

    if(verbosity>=2) t5=get_runtime();

    /* The following computation of the objective function works only */
    /* relative to the active variables */
    if(verbosity>=3) {
      criterion=compute_objective_function(a,lin,label,active2dnum);
      printf("Objective function (over active variables): %.16f\n",criterion);
      fflush(stdout); 
    }

    for(jj=0;working2dnum[jj]>=0;jj++) {
      i=working2dnum[jj]; 
      a_old[i]=a[i];
    }

    retrain=check_optimality(model,label,a,lin,totdoc,svm_c,&maxdiff,
			     &misclassified,inconsistent,active2dnum,
			     last_suboptimal_at,iteration);
    if(verbosity>=2) {
      t6=get_runtime();
      time_select+=t1-t0;
      time_kernel+=t2-t1;
      time_opti+=t3-t2;
      time_update+=t4-t3;
      time_model+=t5-t4;
      time_check+=t6-t5;
    }

    if((!retrain) && (inactivenum>0) && (!skip_final_opt_check)) { 
      if(verbosity>=1) {
	if(verbosity==1) {
	  printf("\n");
	}
	printf(" Checking optimality of inactive variables..."); 
	fflush(stdout);
      }
      t1=get_runtime();
      reactivate_inactive_examples(label,a,a_history,lin,ub_lin,totdoc,
				   iteration,svm_c,inconsistent,active,
				   inactive_since,deactnum,docs,
				   kernel_parm,model,aicache);
      /* Update to new active variables. */
      activenum=compute_index(active,totdoc,active2dnum);
      inactivenum=totdoc-activenum;
      retrain=check_optimality(model,label,a,lin,totdoc,svm_c,&maxdiff,
			       &misclassified,
			       inconsistent,active2dnum,last_suboptimal_at,
			       iteration);
      time_shrink+=get_runtime()-t1;
      if(verbosity>=1) {
	printf("done.\n");  fflush(stdout);
        printf(" Number of inactive variables = %ld\n",inactivenum);
      }		  
    }

    if(verbosity>=2) {
      printf(" => (%ld misclassified, %ld SV (%ld BSV), max violation=%.5f)\n",
	     misclassified,supvecnum,at_upper_bound,maxdiff); 
      fflush(stdout);
    }
    if(verbosity>=3) {
      printf("\n");
    }

    if((!retrain) && remove_inconsistent) {
      if(verbosity>=1) {
	printf(" Moving BSVs to inconsistent examples...");
	fflush(stdout);
      }
      retrain=identify_inconsistent(a,totdoc,svm_c,&inconsistentnum,
				    inconsistent);
      if(verbosity>=1) {
	printf("done.\n");
	if(retrain) {
	  printf(" Now %ld inconsistent examples.\n",inconsistentnum);
	}
      }
    }
    
    if((iteration % 10) == 0) {
      activenum=shrink_problem(svm_iter_to_shrink,last_suboptimal_at,active,
			       inactive_since,active2dnum,&deactnum,
			       iteration,totdoc,max((long)(activenum/10),100),
			       a,a_history,inconsistent);
      inactivenum=totdoc-activenum;
      if((supvecnum>max_kernel_cache_elems)
	 && ((kernel_cache_activenum-activenum)>max((long)(activenum/10),500))) {
	kernel_cache_shrink(totdoc,max((long)(activenum/10),500),active); 
      }
    }
  } /* end of loop */

  if(verbosity>=1) {
    if(verbosity==1) printf("done. (%ld iterations)\n",iteration);
    printf("Optimization finished (%ld misclassified, maxdiff=%.5f).\n",
	   misclassified,maxdiff); 
    runtime_end=get_runtime();
    if(verbosity>=2) {
      printf("Runtime in cpu-minutes: %.2f (%.2f%% for kernel/%.2f%% for optimizer/%.2f%% for final/%.2f%% for update/%.2f%% for model/%.2f%% for check/%.2f%% for select)\n",
	     ((float)runtime_end-(float)runtime_start)/6000.0,
	     (100.0*time_kernel)/(float)(runtime_end-runtime_start),
	     (100.0*time_opti)/(float)(runtime_end-runtime_start),
	     (100.0*time_shrink)/(float)(runtime_end-runtime_start),
	     (100.0*time_update)/(float)(runtime_end-runtime_start),
	     (100.0*time_model)/(float)(runtime_end-runtime_start),
	     (100.0*time_check)/(float)(runtime_end-runtime_start),
	     (100.0*time_select)/(float)(runtime_end-runtime_start));
    }
    else {
      printf("Runtime in cpu-minutes: %.2f\n",
	     (runtime_end-runtime_start)/6000.0);
    }
    if(remove_inconsistent) {
      printf("Number of SV: %ld (plus %ld inconsistent examples)\n",
	     supvecnum,inconsistentnum);
    }
    else {
      upsupvecnum=0;
      for(i=1;i<model->sv_num;i++) {
	if(fabs(model->alpha[i]) >= (svm_c-epsilon_a)) upsupvecnum++;
      }
      printf("Number of SV: %ld (including %ld at upper bound)\n",
	     supvecnum,upsupvecnum);
    }

    if((verbosity>=1) && (!skip_final_opt_check)) {
      model_length=0; 
      for(i=0;i<totdoc;i++) {
	model_length+=a[i]*label[i]*lin[i];
      }
      model_length=sqrt(model_length);
      fprintf(stdout,"Estimated VCdim of classifier: VCdim<=%.5f\n",
	      estimate_margin_vcdim(model,model_length,kernel_parm));
    }
    if(verbosity>=1) {
      printf("Number of kernel evaluations: %ld\n",kernel_cache_statistic);
    }
  }

  free(inconsistent);
  free(chosen);
  free(active);
  free(inactive_since);
  free(last_suboptimal_at);
  free(key);
  free(selcrit);
  free(selexam);
  free(a);
  free(a_old);
  free(a_history);
  free(lin);
  free(ub_lin);
  free(aicache);
  free(working2dnum);
  free(active2dnum);
  free(opt_ce);
  free(opt_ce0);
  free(opt_g);
  free(opt_g0);
  free(opt_xinit);
}

double compute_objective_function(a,lin,label,active2dnum)
double *a,*lin;             /* Return value of objective function. */
long *label,*active2dnum;   /* Works only relative to the active variables! */
{
  long i,ii;
  double criterion;
  /* calculate value of objective function */
  criterion=0;
  for(ii=0;active2dnum[ii]>=0;ii++) {
    i=active2dnum[ii];
    criterion=criterion-a[i]+0.5*a[i]*label[i]*lin[i];
  }
  return(criterion);
}

long compute_index(binfeature,range,index)
long *binfeature,range,*index;  /* create an inverted index of binfeature */
{               
  register long i,ii;

  ii=0;
  for(i=0;i<range;i++) {
    if(binfeature[i]) {
      index[ii]=i;
      ii++;
    }
  }
  index[ii]=-1;
  return(ii);
}

void optimize_svm(docs,label,chosen,active2dnum,model,totdoc,working2dnum,
		  varnum,a,ub_lin,svm_c,aicache,kernel_parm)
DOC *docs;                 /* Do optimization on the working set. */
long *label,*chosen,*active2dnum;
MODEL *model;
long totdoc,*working2dnum,varnum;
double *a,*ub_lin,svm_c;
CFLOAT *aicache;
KERNEL_PARM *kernel_parm;
{
    long i;
    double *a_v;

    compute_matrices_for_optimization(docs,label,chosen,active2dnum,
				      working2dnum,model,a,ub_lin,varnum,
				      totdoc,svm_c,aicache,kernel_parm);

    if(verbosity>=3) {
      printf("Running optimizer..."); fflush(stdout);
    }
    a_v=optimize_qp();      /* call the qp-subsolver */
    if(verbosity>=3) {         
      printf("done\n");
    }

    for(i=0;i<varnum;i++) {
      if(a_v[i]>(0+epsilon_a)) {
	a[working2dnum[i]]=a_v[i];
      }
      else {
	a[working2dnum[i]]=0;
      }
    }
}

void compute_matrices_for_optimization(docs,label,chosen,active2dnum,key,
				       model,a,ub_lin,varnum,totdoc,svm_c,
				       aicache,kernel_parm)
DOC *docs;
long *label,*chosen,*active2dnum,*key;
MODEL *model;
double *a,*ub_lin;
long varnum,totdoc;
double svm_c;
CFLOAT *aicache;
KERNEL_PARM *kernel_parm;
{
  /* writes to the global variables opt_ce, opt_ce0, opt_g, opt_g0 */

  register long ki,i,j,dnum;
  register double qbn,yn;

  if(verbosity>=3) {
    fprintf(stdout,"Computing qp-matrices (type %ld kernel [degree %ld, rbf_gamma %f, coef_lin %f, coef_const %f])...",kernel_parm->kernel_type,kernel_parm->poly_degree,kernel_parm->rbf_gamma,kernel_parm->coef_lin,kernel_parm->coef_const); 
    fflush(stdout);
  }
  
  opt_n=varnum;

  /* Compute the matrix for equality constraints */
  for(i=0;i<varnum;i++) {
    opt_ce[i]=label[key[i]];
  }
  opt_ce0[0]=0; /* compute the constant */
  for(j=1;j<model->sv_num;j++) { /* start at 1 */
    if(!chosen[(model->supvec[j])->docnum]) {
      opt_ce0[0]+=model->alpha[j];
    }
  }
  yn=-opt_ce0[0];

  for(i=0;i<varnum;i++) {
    ki=key[i];
    get_kernel_row(docs,ki,totdoc,active2dnum,aicache,kernel_parm);

    /* compute linear part of objective function */
    qbn=ub_lin[ki];
    for(j=1;j<model->sv_num;j++) { /* start at 1 */
      dnum=(model->supvec[j])->docnum;
      if((!chosen[dnum])
	 && (a[dnum]<=(svm_c-epsilon_a))) { 
	qbn+=aicache[dnum]*model->alpha[j];
      }
      else if((chosen[dnum]) 
	      && (a[dnum]>(svm_c-epsilon_a))) {
	qbn-=aicache[dnum]*model->alpha[j];
      }
    }
    opt_g0[i]=-1+qbn*label[key[i]];
      
    /* compute quadratic part of objective function */
    for(j=i;j<varnum;j++) {
      opt_g[varnum*i+j]=label[ki]*label[key[j]]*aicache[key[j]];
    }
    if(verbosity>=3) {
      if(i % 20 == 0) {
	fprintf(stdout,"%ld..",i); fflush(stdout);
      }
    }
  }

  /* assure starting at feasible point */
  for(i=0;i<varnum;i++) {
    opt_xinit[i]=a[key[i]];
  }

  if(verbosity>=3) {
    fprintf(stdout,"done\n");
  }
}

long calculate_svm_model(docs,label,a,lin,svm_c,chosen,totdoc,
			 at_upper_bound,model,kernel_parm)
DOC *docs;            /* Compute decision function based on current values */
long *label;          /* of alpha. */
double *a,*lin,svm_c;
long *chosen,totdoc,*at_upper_bound;
MODEL *model;
KERNEL_PARM *kernel_parm;
{
  long supvecnum,supvecnum_old=0;
  long i,b_avg_count,first_minlin,first_maxlin;
  double maxlin=0,minlin=0;

  if(verbosity>=3) {
    printf("Calculating model..."); fflush(stdout);
  }
  first_minlin=1;
  first_maxlin=1;
  supvecnum=0;
  (*at_upper_bound)=0;
  b_avg_count=0;
  model->b=0;	       
  model->supvec[0]=0;  /* element 0 reserved and empty for now */
  model->alpha[0]=0;
  supvecnum++;

  for(i=0;i<totdoc;i++) {
    if(a[i]>(0+epsilon_a)) {		/* support vector */
      if((a[i]>(svm_c-epsilon_a))) { 
	(*at_upper_bound)++;
      }
      model->supvec[supvecnum]=&(docs[i]);
      model->alpha[supvecnum]=a[i]*label[i];
      supvecnum++;
      if(a[i]<(svm_c-epsilon_a)) {           /* calculate average b */
	model->b+=(-(double)label[i]+lin[i]);
	b_avg_count++;
      }
      else { /* need maxlin and minlin in case all alphas are at upper bound */
	if((label[i]==1) && ((maxlin<lin[i]) || (first_maxlin))) {
	  first_maxlin=0;
	  maxlin=lin[i];
	}
	else if((label[i]==-1) && ((minlin>lin[i]) || (first_minlin))) {
	  first_minlin=0;
	  minlin=lin[i];
	}
      }
      if(verbosity>=3) {
	if((supvecnum % 100 == 0) && (supvecnum>supvecnum_old)) {
	  supvecnum_old=supvecnum;
	  printf("%ld..",supvecnum); fflush(stdout);
	}
      }
    }
  }
  model->sv_num=supvecnum;
  if(b_avg_count) {             /* now calculate threshold b */
    model->b = model->b / (double)b_avg_count;
  }
  else { /* Warning: No variable not at bound => might need to set C higher! */
    model->b = (maxlin+minlin) / 2.0; /* just pick some b so that KT-Cond */
  }                                   /* can be fulfilled */
  if(verbosity>=3) {
    printf("done (%ld support vectors)\n",supvecnum-1);
  }
  return(supvecnum-1); /* have to substract one, since element 0 is empty */
}


long check_optimality(model,label,a,lin,totdoc,svm_c,maxdiff,
		      misclassified,inconsistent,
		      active2dnum,last_suboptimal_at,iteration)
MODEL *model;            /* Check KT-conditions */
long *label;
double *a,*lin;
long totdoc;
double svm_c,*maxdiff;
long *misclassified;
long *inconsistent,*active2dnum,*last_suboptimal_at,iteration;
{
  long i,ii,retrain;
  double dist,diff;

  epsilon_shrink=epsilon_shrink*0.7+(*maxdiff)*0.3; 
  retrain=0;
  (*maxdiff)=0;
  (*misclassified)=0;
  for(ii=0;active2dnum[ii]>=0;ii++) {
    i=active2dnum[ii];
    if((!inconsistent[i]) && label[i]) {
      dist=(lin[i]-model->b)*(double)label[i];/* 'distance' from hyperplane*/
      diff=fabs(dist-1.0);
      if(dist <= 0) {       
	(*misclassified)++;
      }
      if(diff>(*maxdiff)) { /* largest violation */
	if((a[i]<(svm_c-epsilon_a)) && (dist < 1)) {
	  (*maxdiff)=diff;
	}
	if((a[i]>(0+epsilon_a)) && (dist > 1)) {
	  (*maxdiff)=diff;
	}
      }

      /* Count how long a variable was at lower/upper bound (and optimal).*/
      /* Variables, which were at the bound and optimal for a long */
      /* time are unlikely to become support vectors. In case our */
      /* cache is filled up, those variables are excluded to save */
      /* kernel evaluations. (See chapter 'Shrinking').*/ 
      if((a[i]>=(0+epsilon_a)) && (a[i]<=(svm_c-epsilon_a))) { 
	last_suboptimal_at[i]=iteration;         /* not at bound */
      }
      else if((a[i]<(0+epsilon_a)) && (dist < (1+epsilon_shrink))) {
	last_suboptimal_at[i]=iteration;         /* not likely optimal */
      }
      else if((a[i]>(svm_c-epsilon_a)) && (dist > (1-epsilon_shrink)))  { 
	last_suboptimal_at[i]=iteration;         /* not likely optimal */
      }
    }   
  }
  if((!retrain) && ((*maxdiff) > epsilon_crit)) {  /* termination criterion */
    retrain=1;
  }
  return(retrain);
}

long identify_inconsistent(a,totdoc,svm_c,inconsistentnum,
			   inconsistent)
double *a;
long totdoc;
double svm_c;
long *inconsistentnum,*inconsistent;
{
  long i,retrain;

  /* Throw out examples with multipliers at upper bound. This */
  /* corresponds to the -i 1 option. */
  /* ATTENTION: this is just a heuristic for finding a close */
  /*            to minimum number of examples to exclude to */
  /*            make the problem separable with desired margin */
  retrain=0;
  for(i=0;i<totdoc;i++) {
    if((!inconsistent[i]) && (a[i]>(svm_c-epsilon_a))) { 
	(*inconsistentnum)++;
	inconsistent[i]=1;  /* never choose again */
	retrain=2;          /* start over */
	if(verbosity>=3) {
	  printf("inconsistent(%ld)..",i); fflush(stdout);
	}
    }
  }
  return(retrain);
}

void update_linear_component(docs,label,active2dnum,a,a_old,working2dnum,
			     totdoc,kernel_parm,lin,ub_lin,aicache)
DOC *docs;
long *label,*active2dnum; /* keep track of the linear component lin of the */
double *a,*a_old;        /* gradient etc. by updating based on the change of */
long *working2dnum,totdoc;/* the variables in the current working set */
KERNEL_PARM *kernel_parm;
double *lin,*ub_lin;
CFLOAT *aicache;
{
  register long i,ii,j,jj;
  register double tec;

  for(jj=0;working2dnum[jj]>=0;jj++) {
    i=working2dnum[jj];
    if(a[i] != a_old[i]) {
      get_kernel_row(docs,i,totdoc,active2dnum,aicache,kernel_parm);
      for(ii=0;active2dnum[ii]>=0;ii++) {
	j=active2dnum[ii];
	tec=aicache[j];
	lin[j]+=(((a[i]*tec)-(a_old[i]*tec))*(double)label[i]);
      }
      if((a[i]>(svm_c-epsilon_a)) && (a_old[i]<=(svm_c-epsilon_a))) {
	for(ii=0;active2dnum[ii]>=0;ii++) {
	  j=active2dnum[ii];
	  ub_lin[j]+=a[i]*(double)label[i]*aicache[j];
	}
      }
      else if((a[i]<=(svm_c-epsilon_a)) && (a_old[i]>(svm_c-epsilon_a))) {
	for(ii=0;active2dnum[ii]>=0;ii++) {
	  j=active2dnum[ii];
	  ub_lin[j]-=a_old[i]*(double)label[i]*aicache[j];
	}
      }
    }
  }
}

/*************************** Working set selection ***************************/

long select_next_qp_subproblem_grad(label,a,lin,totdoc,qp_size,svm_c,
				    inconsistent,active2dnum,varnum,
				    working2dnum,selcrit,select,key,chosen)
long *label;
double *a,*lin;      /* Use the feasible direction approach to select the */
long totdoc,qp_size; /* next qp-subproblem  (see section 'Selecting a good */
double svm_c;        /* working set') */
long *inconsistent,*active2dnum,varnum,*working2dnum;
double *selcrit;
long *select,*key,*chosen;
{
  long choosenum,i,j,k,activedoc;
  double s;

  choosenum=0;
  activedoc=0;
  for(i=0;active2dnum[i]>=0;i++) {
    j=active2dnum[i];
    s=-label[j];
    if((!((a[j]<(0+epsilon_a)) && (s<0)))
       && (!((a[j]>(svm_c-epsilon_a)) && (s>0)))
       && (!inconsistent[j]) 
       && (label[j])
       && (!chosen[j])) {
      selcrit[activedoc]=(double)label[j]*(-1.0+(double)label[j]*lin[j]);
      key[activedoc]=j;
      activedoc++;
    }
  }
  select_top_n(selcrit,activedoc,select,(long)(qp_size/2));
  for(k=0;(choosenum<(qp_size/2)) && (k<(qp_size/2)) && (k<activedoc);k++) {
    i=key[select[k]];
    chosen[i]=1;
    choosenum+=1;
    kernel_cache_touch(i); /* make sure it does not get kicked out */
                           /* of cache */
  }

  activedoc=0;
  for(i=0;active2dnum[i]>=0;i++) {
    j=active2dnum[i];
    s=label[j];
    if((!((a[j]<(0+epsilon_a)) && (s<0)))
       && (!((a[j]>(svm_c-epsilon_a)) && (s>0))) 
       && (!inconsistent[j]) 
       && (label[j])
       && (!chosen[j])) {
      selcrit[activedoc]=-(double)(label[j]*(-1.0+(double)label[j]*lin[j]));
      key[activedoc]=j;
      activedoc++;
    }
  }
  select_top_n(selcrit,activedoc,select,(long)(qp_size/2));
  for(k=0;(choosenum<qp_size) && (k<(qp_size/2)) && (k<activedoc);k++) {
    i=key[select[k]];
    chosen[i]=1;
    choosenum+=1;
    kernel_cache_touch(i); /* make sure it does not get kicked out */
                           /* of cache */
  } 
  return(choosenum);
}

long select_next_qp_subproblem_grad_cache(label,a,lin,totdoc,qp_size,svm_c,
					    inconsistent,active2dnum,varnum,
					    working2dnum,selcrit,select,key,
					    chosen)
long *label;
double *a,*lin;      /* Use the feasible direction approach to select the */
long totdoc,qp_size; /* next qp-subproblem  (see chapter 'Selecting a good */
double svm_c;        /* working set') among the variable with cached kernel */
long *inconsistent,*active2dnum,varnum,*working2dnum;
double *selcrit;
long *select,*key,*chosen;
{
  long choosenum,i,j,k,activedoc;
  double s;

  choosenum=0;
  activedoc=0;
  for(i=0;active2dnum[i]>=0;i++) {
    j=active2dnum[i];
    s=-label[j];
    if((!((a[j]<(0+epsilon_a)) && (s<0)))
       && (!((a[j]>(svm_c-epsilon_a)) && (s>0)))
       && (!chosen[j]) 
       && (label[j])
       && (!inconsistent[j]) 
       && (kernel_cache_index[j]>=0))
      {
      selcrit[activedoc]=(double)label[j]*(-1.0+(double)label[j]*lin[j]);
      key[activedoc]=j;
      activedoc++;
    }
  }
  select_top_n(selcrit,activedoc,select,(long)(qp_size/2));
  for(k=0;(choosenum<(qp_size/2)) && (k<(qp_size/2)) && (k<activedoc);k++) {
    i=key[select[k]];
    chosen[i]=1;
    choosenum+=1;
    kernel_cache_touch(i); /* make sure it does not get kicked out */
                           /* of cache */
  }

  activedoc=0;
  for(i=0;active2dnum[i]>=0;i++) {
    j=active2dnum[i];
    s=label[j];
    if((!((a[j]<(0+epsilon_a)) && (s<0)))
       && (!((a[j]>(svm_c-epsilon_a)) && (s>0))) 
       && (!chosen[j]) 
       && (label[j])
       && (!inconsistent[j]) 
       && (kernel_cache_index[j]>=0)) 
      {
      selcrit[activedoc]=-(double)(label[j]*(-1.0+(double)label[j]*lin[j]));
      key[activedoc]=j;
      activedoc++;
    }
  }
  select_top_n(selcrit,activedoc,select,(long)(qp_size/2));
  for(k=0;(choosenum<qp_size) && (k<(qp_size/2)) && (k<activedoc);k++) {
    i=key[select[k]];
    chosen[i]=1;
    choosenum+=1;
    kernel_cache_touch(i); /* make sure it does not get kicked out */
                           /* of cache */
  } 
  return(choosenum);
}

void select_top_n(selcrit,range,select,n)
double *selcrit;
long range,*select,n;
{
  register long i,j;

  for(i=0;(i<n) && (i<range);i++) { /* Initialize with the first n elements */
    for(j=i;j>=0;j--) {
      if((j>0) && (selcrit[select[j-1]]<selcrit[i])){
	select[j]=select[j-1];
      }
      else {
	select[j]=i;
	j=-1;
      }
    }
  }
  for(i=n;i<range;i++) {  
    if(selcrit[i]>selcrit[select[n-1]]) {
      for(j=n-1;j>=0;j--) {
	if((j>0) && (selcrit[select[j-1]]<selcrit[i])) {
	  select[j]=select[j-1];
	}
	else {
	  select[j]=i;
	  j=-1;
	}
      }
    }
  }
}      
      

/******************************** Shrinking  *********************************/

long shrink_problem(svm_iter_to_shrink,last_suboptimal_at,active,
		    inactive_since,active2dnum,deactnum,iteration,totdoc,
		    minshrink,a,a_history,inconsistent)
     /* shrink some variables away */
     /* do the shrinking only if at least minshrink variables can be removed */
long svm_iter_to_shrink,*last_suboptimal_at,*active,*inactive_since;
long *active2dnum,*deactnum,iteration;
long totdoc,minshrink;
double *a,**a_history;
long *inconsistent;
{
  long i,ii,change,activenum;
  double *a_old;
  
  activenum=0;
  change=0;
  for(ii=0;active2dnum[ii]>=0;ii++) {
    i=active2dnum[ii];
    activenum++;
    if(((iteration-last_suboptimal_at[i])>svm_iter_to_shrink) 
       || (inconsistent[i])) {
      change++;
    }
  }
  if(change>=minshrink) { /* shrink only if sufficiently many candidates */
    /* Shrink problem by removing those variables which are */
    /* optimal at a bound for a minimum number of iterations */
    if(verbosity>=2) {
      printf(" Shrinking..."); fflush(stdout);
    }
    a_old=(double *)my_malloc(sizeof(double)*totdoc);
    a_history[(*deactnum)]=a_old;
    for(i=0;i<totdoc;i++) {
      a_old[i]=a[i];
    }
    change=0;
    for(ii=0;active2dnum[ii]>=0;ii++) {
      i=active2dnum[ii];
      if((((iteration-last_suboptimal_at[i])>svm_iter_to_shrink) 
	  || (inconsistent[i]))) {
	active[i]=0;
	inactive_since[i]=(*deactnum);
	change++;
      }
    }
    activenum=compute_index(active,totdoc,active2dnum);
    (*deactnum)++;
    if(verbosity>=2) {
      printf("done.\n"); fflush(stdout);
      printf(" Number of inactive variables = %ld\n",totdoc-activenum);
    }
  }
  return(activenum);
} 

void reactivate_inactive_examples(label,a,a_history,lin,ub_lin,totdoc,
				  iteration,svm_c,inconsistent,active,
				  inactive_since,deactnum,docs,kernel_parm,
				  model,aicache)
long *label;                         /* Make all variables active again */
double *a,**a_history,*lin,*ub_lin;  /* which had been removed by shrinking. */
long totdoc,iteration;               /* Computes lin and ub_lin for those */
double svm_c;                        /* variables from scratch. */
long *inconsistent,*active,*inactive_since,deactnum;
DOC *docs;
KERNEL_PARM *kernel_parm;
MODEL *model;
CFLOAT *aicache;
{
  register long i,j,ii,jj,t,*changed2dnum,*inactive2dnum;
  long *changed,*inactive;
  register double kernel_val,*a_old,dist;

  changed=(long *)my_malloc(sizeof(long)*totdoc);
  changed2dnum=(long *)my_malloc(sizeof(long)*(totdoc+1));
  inactive=(long *)my_malloc(sizeof(long)*totdoc);
  inactive2dnum=(long *)my_malloc(sizeof(long)*(totdoc+1));
  for(t=deactnum-1;(t>=0) && a_history[t];t--) {
    if(verbosity>=2) {
      printf("%ld..",t); fflush(stdout);
    }
    a_old=a_history[t];    
    for(i=0;i<totdoc;i++) {
      inactive[i]=((!active[i]) 
		  && (inactive_since[i] == t));
      changed[i]= (a[i] != a_old[i]);
    }
    compute_index(inactive,totdoc,inactive2dnum);
    compute_index(changed,totdoc,changed2dnum);

    for(ii=0;changed2dnum[ii]>=0;ii++) {
      i=changed2dnum[ii];
      get_kernel_row(docs,i,totdoc,inactive2dnum,aicache,kernel_parm);
      for(jj=0;inactive2dnum[jj]>=0;jj++) {
	j=inactive2dnum[jj];
	kernel_val=aicache[j];
	lin[j]+=(((a[i]*kernel_val)-(a_old[i]*kernel_val))*(double)label[i]);
	if((a[i]>(svm_c-epsilon_a)) && (a_old[i]<=(svm_c-epsilon_a))) {
	  ub_lin[j]+=a[i]*(double)label[i]*kernel_val;
	}
	else if((a[i]<=(svm_c-epsilon_a)) && (a_old[i]>(svm_c-epsilon_a))) {
	  ub_lin[j]-=a_old[i]*(double)label[i]*kernel_val;
	}
      }
    }
  }
  for(i=0;i<totdoc;i++) {
    inactive_since[i]=deactnum-1;
    if(!inconsistent[i]) {
      dist=(lin[i]-model->b)*(double)label[i];
      if((a[i]>=(0+epsilon_a)) && (a[i]<=(svm_c-epsilon_a))) { 
	active[i]=1;                            /* not at bound */
      }
      else if((a[i]<(0+epsilon_a)) && (dist < (1+epsilon_shrink))) {
	active[i]=1;
      }
      else if((a[i]>(svm_c-epsilon_a))&& (dist > (1-epsilon_shrink))) {
	active[i]=1;
      }
    }
  }
  for(i=0;i<totdoc;i++) {
    (a_history[deactnum-1])[i]=a[i];
  }
  for(t=deactnum-2;(t>=0) && a_history[t];t--) {
      free(a_history[t]);
      a_history[t]=0;
  }
  free(changed);
  free(changed2dnum);
  free(inactive);
  free(inactive2dnum);
}

/****************************** Cache handling *******************************/

void get_kernel_row(docs,docnum,totdoc,active2dnum,buffer,kernel_parm)
DOC *docs;          /* Get's a row of the matrix of kernel values */
long docnum,totdoc; /* This matrix has the same form as the Hessian, */ 
long *active2dnum;  /* just that the elements are not multiplied by */
CFLOAT *buffer;     /* y_i * y_j * a_i * a_j */
KERNEL_PARM *kernel_parm; /* Takes the values from the cache if available. */
{
  register long i,j,start;
  DOC *ex;

  ex=&(docs[docnum]);
  if(kernel_cache_index[docnum] != -1) {
    kernel_cache_lru[kernel_cache_index[docnum]]=kernel_cache_time; /* lru */
    start=kernel_cache_activenum*kernel_cache_index[docnum];
    for(i=0;active2dnum[i]>=0;i++) {
      j=active2dnum[i];
      if(kernel_cache_totdoc2active[j] >= 0) {
	buffer[j]=kernel_cache_buffer[start+kernel_cache_totdoc2active[j]];
      }
      else {
	buffer[j]=(CFLOAT)kernel(kernel_parm,ex,&(docs[j]));
      }
    }
  }
  else {
    for(i=0;active2dnum[i]>=0;i++) {
      j=active2dnum[i];
      buffer[j]=(CFLOAT)kernel(kernel_parm,ex,&(docs[j]));
    }
  }
}

void cache_kernel_evals(docs,key,varnum,kernel_parm)
DOC *docs;  /* Fills cache for the rows in key */
long *key;
long varnum;
KERNEL_PARM *kernel_parm;
{
  register DOC *ex;
  register long i,j,k,l,m;
  register CFLOAT *cache;

  for(i=0;i<varnum;i++) {  /* fill up kernel cache */
    m=key[i];
    if(!kernel_cache_check(m)) {  /* not cached yet*/
      cache = kernel_cache_clean_and_malloc(m);
      if(cache) {
	l=kernel_cache_totdoc2active[m];
	ex=&(docs[m]);
	for(j=0;j<kernel_cache_activenum;j++) {  /* fill cache */
	  k=kernel_cache_active2totdoc[j];
	  if((kernel_cache_index[k] != -1) && (l != -1) && (k != m)) {
	    cache[j]=kernel_cache_buffer[kernel_cache_activenum
					*kernel_cache_index[k]+l];
	  }
	  else {
	    cache[j]=kernel(kernel_parm,ex,&(docs[k]));
	  } 
	}
      }
      else {
	perror("Error: Kernel cache full! => increase cache size");
      }
    }
  }
}

 
void kernel_cache_shrink(totdoc,numshrink,after)
long totdoc,numshrink,*after; /* remove numshrink columns in the cache */
{                           /* which correspond to examples marked  */
  register long i,j,jj,from=0,to=0,scount;     /* 0 in after. */
  long *keep;

  if(verbosity>=2) {
    printf(" Reorganizing cache..."); fflush(stdout);
  }

  keep=(long *)my_malloc(sizeof(long)*totdoc);
  for(j=0;j<totdoc;j++) {
    keep[j]=1;
  }
  scount=0;
  for(jj=0;(jj<kernel_cache_activenum) && (scount<numshrink);jj++) {
    j=kernel_cache_active2totdoc[jj];
    if(!after[j]) {
      scount++;
      keep[j]=0;
    }
  }

  for(i=0;i<max_kernel_cache_elems;i++) {
    for(jj=0;jj<kernel_cache_activenum;jj++) {
      j=kernel_cache_active2totdoc[jj];
      if(!keep[j]) {
	from++;
      }
      else {
	kernel_cache_buffer[to]=kernel_cache_buffer[from];
	to++;
	from++;
      }
    }
  }

  kernel_cache_activenum=0;
  for(j=0;j<totdoc;j++) {
    if((keep[j]) && (kernel_cache_totdoc2active[j] != -1)) {
      kernel_cache_active2totdoc[kernel_cache_activenum]=j;
      kernel_cache_totdoc2active[j]=kernel_cache_activenum;
      kernel_cache_activenum++;
    }
    else {
      kernel_cache_totdoc2active[j]=-1;
    }
  }

  max_kernel_cache_elems=(long)(kernel_cache_buffsize/kernel_cache_activenum);
  if(max_kernel_cache_elems>totdoc) {
    max_kernel_cache_elems=totdoc;
  }

  free(keep);

  if(verbosity>=2) {
    printf("done.\n"); fflush(stdout);
    printf(" Cache-size in rows = %ld\n",max_kernel_cache_elems);
  }
}

void kernel_cache_init(totdoc,buffsize)
long totdoc,buffsize;
{
  long i;

  kernel_cache_index = (long *)my_malloc(sizeof(long)*totdoc);
  kernel_cache_occu = (long *)my_malloc(sizeof(long)*totdoc);
  kernel_cache_lru = (long *)my_malloc(sizeof(long)*totdoc);
  kernel_cache_invindex = (long *)my_malloc(sizeof(long)*totdoc);
  kernel_cache_active2totdoc = (long *)my_malloc(sizeof(long)*totdoc);
  kernel_cache_totdoc2active = (long *)my_malloc(sizeof(long)*totdoc);
  kernel_cache_buffer = (CFLOAT *)my_malloc(buffsize*1024*1024);

  kernel_cache_buffsize=(long)(buffsize*1024*1024/sizeof(CFLOAT));

  max_kernel_cache_elems=(long)(kernel_cache_buffsize/totdoc);
  if(max_kernel_cache_elems>totdoc) {
    max_kernel_cache_elems=totdoc;
  }

  if(verbosity>=2) {
    printf(" Cache-size in rows = %ld\n",max_kernel_cache_elems);
    printf(" Kernel evals so far: %ld\n",kernel_cache_statistic);    
  }

  kernel_cache_elems=0;   /* initialize cache */
  for(i=0;i<totdoc;i++) {
    kernel_cache_index[i]=-1;
  }
  for(i=0;i<max_kernel_cache_elems;i++) {
    kernel_cache_occu[i]=0;
    kernel_cache_invindex[i]=-1;
  }

  kernel_cache_activenum=totdoc;;
  for(i=0;i<totdoc;i++) {
      kernel_cache_active2totdoc[i]=i;
      kernel_cache_totdoc2active[i]=i;
  }
} 

void kernel_cache_cleanup()
{
  free(kernel_cache_index);
  free(kernel_cache_occu);
  free(kernel_cache_lru);
  free(kernel_cache_invindex);
  free(kernel_cache_active2totdoc);
  free(kernel_cache_totdoc2active);
  free(kernel_cache_buffer);
}

long kernel_cache_malloc() 
{
  long i;

  if(kernel_cache_elems < max_kernel_cache_elems) {
    for(i=0;i<max_kernel_cache_elems;i++) {
      if(!kernel_cache_occu[i]) {
	kernel_cache_occu[i]=1;
	kernel_cache_elems++;
	return(i);
      }
    }
  }
  return(-1);
}

void kernel_cache_free(i)
long i;
{
  kernel_cache_occu[i]=0;
  kernel_cache_elems--;
}

long kernel_cache_free_lru()          /* remove least recently used cache */
{                                     /* element */
  register long k,least_elem=-1,least_time;

  least_time=kernel_cache_time+1;
  for(k=0;k<max_kernel_cache_elems;k++) {
    if(kernel_cache_invindex[k] != -1) {
      if(kernel_cache_lru[k]<least_time) {
	least_time=kernel_cache_lru[k];
	least_elem=k;
      }
    }
  }
  if(least_elem != -1) {
    kernel_cache_free(least_elem);
    kernel_cache_index[kernel_cache_invindex[least_elem]]=-1;
    kernel_cache_invindex[least_elem]=-1;
    return(1);
  }
  return(0);
}


CFLOAT *kernel_cache_clean_and_malloc(docnum)
long docnum;  /* Get a free cache entry. In case cache is full, the lru */
{             /* element is removed. */
  long result;
  if((result = kernel_cache_malloc()) == -1) {
    if(kernel_cache_free_lru()) {
      result = kernel_cache_malloc();
    }
  }
  kernel_cache_index[docnum]=result;
  if(result == -1) {
    return(0);
  }
  kernel_cache_invindex[result]=docnum;
  kernel_cache_lru[kernel_cache_index[docnum]]=kernel_cache_time; /* lru */
  return((CFLOAT *)((long)kernel_cache_buffer
		    +(kernel_cache_activenum*sizeof(CFLOAT)*
		      kernel_cache_index[docnum])));
}

long kernel_cache_touch(docnum)
long docnum; /* Update lru time to avoid removal from cache. */
{
  if(kernel_cache_index[docnum] != -1) {
    kernel_cache_lru[kernel_cache_index[docnum]]=kernel_cache_time; /* lru */
    return(1);
  }
  return(0);
}
  
long kernel_cache_check(docnum)
long docnum; /* Is that row cached? */
{
  return(kernel_cache_index[docnum] != -1);
}
  
/************************** Compute VCdim bound ******************************/

double estimate_margin_vcdim(model,w,kernel_parm) 
MODEL *model;
double w; /* optional: length of model vector in feature space */
KERNEL_PARM *kernel_parm;
{
  double h,R;

  /* follows chapter 5.6.4 in [Vapnik/95] */

  if(w<0) {
    w=model_length_s(model,kernel_parm);
  }
  R=estimate_sphere(model,kernel_parm); 
  if(verbosity>=1) {
    fprintf(stdout,"Norm of weight vector: |w|=%.5f\n",w);
  }
  h = w*w * R*R +1; 
  return(h);
}

double estimate_sphere(model,kernel_parm) 
MODEL *model;             /* Approximates the radius of the ball containing */
KERNEL_PARM *kernel_parm; /* the support vectors by bounding it with the */
{                         /* length of the longest support vector. This is */
  register long j;        /* pretty good for text categorization, since all */
  double xlen,maxxlen=0;  /* documents have feature vectors of length 1. It */
  DOC nulldoc;            /* assumes that the center of the ball is at the */
  WORD nullword;          /* origin of the space. */

  if(verbosity>=1) {
    fprintf(stdout,"Estimating radius R of sphere..."); fflush(stdout);
  }
  nullword.wnum=0;
  nulldoc.words=&nullword;
  nulldoc.twonorm_sq=0;
  nulldoc.docnum=-1;

  for(j=1;j<model->sv_num;j++) {
    xlen=sqrt(kernel(kernel_parm,model->supvec[j],model->supvec[j])
	      -2*kernel(kernel_parm,model->supvec[j],&nulldoc)
	      +kernel(kernel_parm,&nulldoc,&nulldoc));
    if(xlen>maxxlen) {
      maxxlen=xlen;
    }
  }
  if(verbosity>=1) {
    fprintf(stdout,"done (R <= longest support vector = %.5f)\n",maxxlen);
  }
  
  return(maxxlen);
}

/****************************** IO-handling **********************************/

void write_model(modelfile,model,kernel_parm)
char *modelfile;
MODEL *model;
KERNEL_PARM *kernel_parm;
{
  FILE *modelfl;
  long j,i;

  if(verbosity>=1) {
    printf("Writing model file..."); fflush(stdout);
  }
  if ((modelfl = fopen (modelfile, "w")) == NULL)
  { perror (modelfile); exit (1); }
  fprintf(modelfl,"SVM-light Version %s\n",VERSION);
  fprintf(modelfl,"%ld\n",kernel_parm->kernel_type);
  fprintf(modelfl,"%ld\n",kernel_parm->poly_degree);
  fprintf(modelfl,"%.8g\n",kernel_parm->rbf_gamma);
  fprintf(modelfl,"%.8g\n",kernel_parm->coef_lin);
  fprintf(modelfl,"%.8g\n",kernel_parm->coef_const);
  fprintf(modelfl,"%s\n",kernel_parm->custom);
 
  fprintf(modelfl,"%ld\n",model->sv_num);
  fprintf(modelfl,"%.8g\n",model->b);
  for(i=1;i<model->sv_num;i++) {
    fprintf(modelfl,"%.32g ",model->alpha[i]);
    for (j=0; ((model->supvec[i])->words[j]).wnum; j++) {
      fprintf(modelfl,"%ld:%.8g ",
	      (long)((model->supvec[i])->words[j]).wnum,
	      (double)(((model->supvec[i])->words[j]).weight));
    }
    fprintf(modelfl,"\n");
  }
  fclose(modelfl);
  if(verbosity>=1) {
    printf("done\n");
  }
}

void read_input_parameters(argc,argv,docfile,modelfile,verbosity,
			   remove_inconsistent,skip_final_opt_check,
			   svm_maxqpsize,svm_iter_to_shrink,kernel_cache_size,
			   svm_c,epsilon_crit,kernel_parm)
int argc;
char *argv[],*docfile,*modelfile;
long *verbosity,*remove_inconsistent,*skip_final_opt_check,*svm_maxqpsize;
long *svm_iter_to_shrink,*kernel_cache_size;
double *svm_c,*epsilon_crit;
KERNEL_PARM *kernel_parm;
{
  long i;
  
  /* set default */
  strcpy (modelfile, "svm_model");
  (*verbosity)=2;
  (*remove_inconsistent)=0;
  (*skip_final_opt_check)=0;
  (*svm_maxqpsize)=10;
  (*svm_iter_to_shrink)=100;
  (*kernel_cache_size)=40;
  (*svm_c)=1000;
  (*epsilon_crit)=0.001;
  kernel_parm->kernel_type=0;
  kernel_parm->poly_degree=3;
  kernel_parm->rbf_gamma=1.0;
  kernel_parm->coef_lin=1;
  kernel_parm->coef_const=1;
  strcpy(kernel_parm->custom,"empty");

  for(i=1;(i<argc) && ((argv[i])[0] == '-');i++) {
    switch ((argv[i])[1]) 
      { 
      case '?': print_help(); exit(0);
      case 'v': i++; (*verbosity)=atol(argv[i]); break;
      case 'i': i++; (*remove_inconsistent)=atol(argv[i]); break;
      case 'f': i++; (*skip_final_opt_check)=!atol(argv[i]); break;
      case 'q': i++; (*svm_maxqpsize)=atol(argv[i]); break;
      case 'h': i++; (*svm_iter_to_shrink)=atol(argv[i]); break;
      case 'm': i++; (*kernel_cache_size)=atol(argv[i]); break;
      case 'c': i++; (*svm_c)=atof(argv[i]); break;
      case 'e': i++; (*epsilon_crit)=atof(argv[i]); break;
      case 't': i++; kernel_parm->kernel_type=atol(argv[i]); break;
      case 'd': i++; kernel_parm->poly_degree=atol(argv[i]); break;
      case 'g': i++; kernel_parm->rbf_gamma=atof(argv[i]); break;
      case 's': i++; kernel_parm->coef_lin=atof(argv[i]); break;
      case 'r': i++; kernel_parm->coef_const=atof(argv[i]); break;
      case 'u': i++; strcpy(kernel_parm->custom,argv[i]); break;
      default: printf("\nUnrecognized option %s!\n\n",argv[i]);
	       print_help();
	       exit(0);
      }
  }
  if(i>=argc) {
    printf("\nNot enough input parameters!\n\n");
    print_help();
    exit(0);
  }
  strcpy (docfile, argv[i]);
  if((i+1)<argc) {
    strcpy (modelfile, argv[i+1]);
  }
  if(((*svm_maxqpsize)<2) || ((*svm_maxqpsize)>nx)) {
    printf("\nMaximum size of QP-subproblems not in valid range: %ld [2,%ld]\n",(*svm_maxqpsize),(long)nx); 
    print_help();
    exit(0);
  }
  if((*svm_iter_to_shrink)<1) {
    printf("\nMaximum number of iterations for shrinking not in valid range: %ld [1,..]\n",(*svm_iter_to_shrink));
    print_help();
    exit(0);
  }
  if((*svm_c)<=0) {
    printf("\nThe C parameter must be greater than zero!\n\n");
    print_help();
    exit(0);
  }
  if((*epsilon_crit)<=0) {
    printf("\nThe epsilon parameter must be greater than zero!\n\n");
    print_help();
    exit(0);
  }
}

void print_help()
{
  printf("\nSVM-light %s: Support Vector Machine, learning module     %s\n",VERSION,VERSION_DATE);
  copyright_notice();
  printf("   usage: svm_learn [options] example_file model_file\n\n");
  printf("options: -?          -> this help\n");
  printf("         -v [0..3]   -> verbosity level (default 2)\n");
  printf("         -i [0,1]    -> remove inconsistent training examples\n");
  printf("                        and retrain (default 0)\n");
  printf("         -q [2..%d] -> maximum size of QP-subproblems (default 10)\n",nx);
  printf("         -h [5..]    -> number of iterations a variable needs to be\n"); 
  printf("                        optimal before considered for shrinking (default 100)\n");
  printf("         -f [0,1]    -> do final optimality check for variables removed\n");
  printf("                        by shrinking. Although this test is usually \n");
  printf("                        positive, there is no guarantee that the optimum\n");
  printf("                        was found if the test is omitted. (default 1)\n");
  printf("         -m [5..]    -> size of cache for kernel evaluations in MB (default 40)\n");
  printf("                        The larger the faster...\n");
  printf("         -c float    -> C: trade-off between training error\n");
  printf("                        and margin (default 1000)\n");
  printf("         -e float    -> eps: Allow that error for termination criterion\n");
  printf("                        [y [w*x+b] - 1] >= eps (default 0.001)\n");
  printf("         -t int      -> type of kernel function:\n");
  printf("                        0: linear\n");
  printf("                        1: polynomial (s a*b+c)^d\n");
  printf("                        2: radial basis function exp(-gamma ||a-b||^2)\n");
  printf("                        3: sigmoid tanh(s a*b + c)\n");
  printf("                        4: user defined kernel from kernel.h\n");
  printf("         -d int      -> parameter d in polynomial kernel\n");
  printf("         -g float    -> parameter gamma in rbf kernel\n");
  printf("         -s float    -> parameter s in sigmoid/poly kernel\n");
  printf("         -r float    -> parameter c in sigmoid/poly kernel\n");
  printf("         -u string   -> parameter of user defined kernel\n\n");
}




